package org.lwjglx.opengl;

import static org.lwjgl.opengl.GL11.GL_FALSE;
import static org.lwjgl.opengl.GL11.GL_TRUE;
import static org.lwjgl.system.MemoryUtil.NULL;
import static org.lwjgl.system.glfw.GLFW.GLFW_OPENGL_DEBUG_CONTEXT;
import static org.lwjgl.system.glfw.GLFW.GLFW_RESIZABLE;
import static org.lwjgl.system.glfw.GLFW.GLFW_VISIBLE;
import static org.lwjgl.system.glfw.GLFW.glfwCreateWindow;
import static org.lwjgl.system.glfw.GLFW.glfwDefaultWindowHints;
import static org.lwjgl.system.glfw.GLFW.glfwDestroyWindow;
import static org.lwjgl.system.glfw.GLFW.glfwGetPrimaryMonitor;
import static org.lwjgl.system.glfw.GLFW.glfwGetVideoMode;
import static org.lwjgl.system.glfw.GLFW.glfwMakeContextCurrent;
import static org.lwjgl.system.glfw.GLFW.glfwPollEvents;
import static org.lwjgl.system.glfw.GLFW.glfwSetWindowPos;
import static org.lwjgl.system.glfw.GLFW.glfwShowWindow;
import static org.lwjgl.system.glfw.GLFW.glfwSwapBuffers;
import static org.lwjgl.system.glfw.GLFW.glfwSwapInterval;
import static org.lwjgl.system.glfw.GLFW.glfwWindowHint;
import static org.lwjgl.system.glfw.GLFW.glfwWindowShouldClose;

import java.nio.ByteBuffer;
import java.nio.IntBuffer;

import org.lwjgl.BufferUtils;
import org.lwjgl.system.glfw.GLFW;
import org.lwjgl.system.glfw.GLFWvidmode;
import org.lwjglx.LWJGLException;
import org.lwjglx.Sys;

public class Display {
	
	private static String windowTitle = "Game";
	
	private static long      window;
	private static org.lwjgl.opengl.GLContext context;
	
	private static DisplayMode mode = new DisplayMode(640, 480);
	
	public static void create() throws LWJGLException {
		org.lwjgl.Sys.touch();
		
		Sys.getVersion(); // init glfw

		long monitor = glfwGetPrimaryMonitor();
		ByteBuffer vidmode = glfwGetVideoMode(monitor);

		int monitorWidth = GLFWvidmode.width(vidmode);
		int monitorHeight = GLFWvidmode.height(vidmode);

		glfwDefaultWindowHints();
		glfwWindowHint(GLFW_VISIBLE, GL_FALSE);
		glfwWindowHint(GLFW_RESIZABLE, GL_FALSE);
		glfwWindowHint(GLFW_OPENGL_DEBUG_CONTEXT, GL_TRUE);

		window = glfwCreateWindow(mode.getWidth(), mode.getHeight(), windowTitle, NULL, NULL);

		glfwSetWindowPos(
			window,
			(monitorWidth - mode.getWidth()) / 2,
			(monitorHeight - mode.getHeight()) / 2
		);

		glfwMakeContextCurrent(window);
		context = org.lwjgl.opengl.GLContext.createFromCurrent();
		
		glfwSwapInterval(1);
		glfwShowWindow(window);
	}
	
	public static org.lwjgl.opengl.GLContext getContext() {
		return context;
	}
	
	/**
	 * Set the window's location. This is a no-op on fullscreen windows or when getParent() != null.
	 * The window is clamped to remain entirely on the screen. If you attempt
	 * to position the window such that it would extend off the screen, the window
	 * is simply placed as close to the edge as possible.
	 * <br><b>note</b>If no location has been specified (or x == y == -1) the window will be centered
	 *
	 * @param new_x The new window location on the x axis
	 * @param new_y The new window location on the y axis
	 */
	public static void setLocation(int new_x, int new_y) {
		System.out.println("TODO: Implement Display.setLocation(int, int)");
	}
	
	public static long getWindow() {
		return window;
	}
	
	public static void update() {
		glfwPollEvents();
		glfwSwapBuffers(window);
	}
	
	public static void destroy() {
		glfwDestroyWindow(window);
		
		/*try {
			glfwTerminate();
		} catch (Throwable t) {
			t.printStackTrace();
		}*/
	}
	
	public static void setDisplayMode(DisplayMode dm) {
		mode = dm;
	}
	
	public static DisplayMode getDisplayMode() {
		return mode;
	}
	
	public static DisplayMode[] getAvailableDisplayModes() throws LWJGLException {
		IntBuffer count = BufferUtils.createIntBuffer(1);
		ByteBuffer modes = GLFW.glfwGetVideoModes(GLFW.glfwGetPrimaryMonitor(), count);

		DisplayMode[] displayModes = new DisplayMode[count.get(0)];

		for (int i = 0; i < count.get(0); i++) {
			modes.position(i * GLFWvidmode.SIZEOF);

			int w = GLFWvidmode.width(modes);
			int h = GLFWvidmode.height(modes);
			int b = GLFWvidmode.redBits(modes) + GLFWvidmode.greenBits(modes)
					+ GLFWvidmode.blueBits(modes);
			int r = GLFWvidmode.refreshRate(modes);

			displayModes[i] = new DisplayMode(w, h, b, r);
		}

		return displayModes;
	}
	
	public static void setTitle(String title) {
		windowTitle = title;
	}
	
	public static boolean isCloseRequested() {
		return glfwWindowShouldClose(window) == GL_TRUE;
	}
	
	public static void setFullscreen(boolean fullscreen) {
		// TODO
	}
	
	/**
	 * An accurate sync method that will attempt to run at a constant frame rate.
	 * It should be called once every frame.
	 *
	 * @param fps - the desired frame rate, in frames per second
	 */
	public static void sync(int fps) {
		Sync.sync(fps);
	}

}
